# Please report

- [ ] version of ABySS with `abyss-pe version`
- [ ] distribution of Linux with `lsb_release -d`

# Assembly error

- [ ] complete `abyss-pe` command line
- [ ] last 20 lines of the output of `abyss-pe`
- [ ] number of sequenced bases
- [ ] estimated genome size and ploidy
- [ ] estimated sequencing depth of coverage

# Build error

Consider installing ABySS using [Linuxbrew](https://linuxbrew.sh) on Linux or [Homebrew](https://brew.sh) on macOS with `brew install abyss`, or using [Bioconda](https://bioconda.github.io) with `conda install abyss`.

- [ ] Have you tried installing ABySS using Brew or Bioconda?
- [ ] version of GCC or compiler with `gcc --version`
- [ ] complete `./configure` command line
- [ ] last 20 lines of the output of `./configure`
- [ ] last 20 lines of the output of `make`
