#ifndef PARALLEL_SEQUENCECOLLECTION_H
#define PARALLEL_SEQUENCECOLLECTION_H 1

#if PAIRED_DBG
# include "PairedDBG/SequenceCollection.h"
#else
# include "Assembly/SequenceCollection.h"
#endif

#endif
