/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 4.0.1
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

package btllib;

public class KmerBloomFilter {
  private transient long swigCPtr;
  protected transient boolean swigCMemOwn;

  protected KmerBloomFilter(long cPtr, boolean cMemoryOwn) {
    swigCMemOwn = cMemoryOwn;
    swigCPtr = cPtr;
  }

  protected static long getCPtr(KmerBloomFilter obj) {
    return (obj == null) ? 0 : obj.swigCPtr;
  }

  @SuppressWarnings("deprecation")
  protected void finalize() {
    delete();
  }

  public synchronized void delete() {
    if (swigCPtr != 0) {
      if (swigCMemOwn) {
        swigCMemOwn = false;
        btllibJNI.delete_KmerBloomFilter(swigCPtr);
      }
      swigCPtr = 0;
    }
  }

  public KmerBloomFilter(long k, long bytes, long hash_num) {
    this(btllibJNI.new_KmerBloomFilter__SWIG_0(k, bytes, hash_num), true);
  }

  public KmerBloomFilter(long k, long bytes) {
    this(btllibJNI.new_KmerBloomFilter__SWIG_1(k, bytes), true);
  }

  public void insert(String seq) {
    btllibJNI.KmerBloomFilter_insert__SWIG_0(swigCPtr, this, seq);
  }

  public void insert(String seq, long seq_len) {
    btllibJNI.KmerBloomFilter_insert__SWIG_1(swigCPtr, this, seq, seq_len);
  }

  public long contains(String seq) {
    return btllibJNI.KmerBloomFilter_contains__SWIG_0(swigCPtr, this, seq);
  }

  public long contains(String seq, long seq_len) {
    return btllibJNI.KmerBloomFilter_contains__SWIG_1(swigCPtr, this, seq, seq_len);
  }

}
