#include "Common/Options.h"

namespace opt {
	/** Colour space sequences */
	bool colourSpace;

	/** MPI rank */
	int rank = -1;

	/** Number of MPI processes */
	int numProc = 1;

	/** Verbose output */
	int verbose;
}
