/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 4.0.1
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

package btllib;

public class DataSink extends DataStream {
  private transient long swigCPtr;

  protected DataSink(long cPtr, boolean cMemoryOwn) {
    super(btllibJNI.DataSink_SWIGUpcast(cPtr), cMemoryOwn);
    swigCPtr = cPtr;
  }

  protected static long getCPtr(DataSink obj) {
    return (obj == null) ? 0 : obj.swigCPtr;
  }

  @SuppressWarnings("deprecation")
  protected void finalize() {
    delete();
  }

  public synchronized void delete() {
    if (swigCPtr != 0) {
      if (swigCMemOwn) {
        swigCMemOwn = false;
        btllibJNI.delete_DataSink(swigCPtr);
      }
      swigCPtr = 0;
    }
    super.delete();
  }

  public DataSink(String path, boolean append) {
    this(btllibJNI.new_DataSink__SWIG_0(path, append), true);
  }

  public DataSink(String path) {
    this(btllibJNI.new_DataSink__SWIG_1(path), true);
  }

}
