/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 4.0.1
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

package btllib;

public class SWIGTYPE_p_bool {
  private transient long swigCPtr;

  protected SWIGTYPE_p_bool(long cPtr, @SuppressWarnings("unused") boolean futureUse) {
    swigCPtr = cPtr;
  }

  protected SWIGTYPE_p_bool() {
    swigCPtr = 0;
  }

  protected static long getCPtr(SWIGTYPE_p_bool obj) {
    return (obj == null) ? 0 : obj.swigCPtr;
  }
}

