#ifndef COMMON_OPTIONS_H
#define COMMON_OPTIONS_H 1

/**
 * Global variables that are mostly constant for the duration of the
 * execution of the program.
 */
namespace opt {
	extern bool colourSpace;
	extern int numProc;
	extern int rank;
	extern int verbose;
}

#endif
